"""Shared pytest fixtures."""
from typing import Any, Dict

import pytest

from cliconfig.process_routines import Processing


class ProcessAdd1(Processing):
    """Add 1 to values with tag "@add+1"."""

    # pylint: disable=unused-argument
    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-merge processing."""
        items = list(flat_dict.items())
        for key, value in items:
            if key.endswith("@add+1"):
                flat_dict[key[:-6]] = value + 1
                del flat_dict[key]
        return flat_dict

    def presave(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list
    ) -> Dict[str, Any]:
        """Pre-save processing."""
        return self.premerge(flat_dict, processing_list)


class ProcessKeep(Processing):
    """Prevent a value from being changed after the merge."""

    def __init__(self) -> None:
        super().__init__()
        self.keep_vals: Dict[str, Any] = {}

    # pylint: disable=unused-argument
    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-merge processing."""
        items = list(flat_dict.items())
        for key, value in items:
            if key.endswith("@keep"):
                new_key = key[:-5]
                flat_dict[new_key] = value
                del flat_dict[key]
                self.keep_vals[new_key] = value
        return flat_dict

    # pylint: disable=unused-argument
    def postmerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list, # noqa
    ) -> Dict[str, Any]:
        """Post-merge processing."""
        for key, value in self.keep_vals.items():
            flat_dict[key] = value
        self.keep_vals = {}
        return flat_dict

    # pylint: disable=unused-argument
    def postload(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Post-load processing."""
        for key, value in self.keep_vals.items():
            flat_dict[key] = value
        self.keep_vals = {}
        return flat_dict


@pytest.fixture()
def process_add1() -> ProcessAdd1:
    """Return a processing object that adds 1 on tag "@add+1"."""
    return ProcessAdd1()


@pytest.fixture()
def process_keep() -> ProcessKeep:
    """Return a processing object that keep a value unchanged after the merge."""
    return ProcessKeep()



"""Example of useful processing classes.

Classes to apply pre-merge, post-merge, pre-save and post-load modifications
to dict with processing routines (found in cliconfig.process_routines).
"""
from typing import Any, Dict

from cliconfig.process_routines import Processing, merge_flat_paths_processing


class ProcessMerge(Processing):
    """Merge dict just in time with '@merge_after' and '@merge_before' tags.

    Tag your key with '@merge_after' or '@merge_before' to load the dict
    corresponding to a path and merge it just before or after the current
    dicturation. The processing is a pre-merge processing only and occurs before
    most of the other processing.
    Pre-merge order: -20.0

    Examples
    --------
    .. code-block:: yaml

        --- # config1.yaml
        a:
          b: 1
          b_path@merge_after: dict2.yaml
        --- # config2.yaml
        a.b: 2

    Before merging, the config1 is interpreted as the dict
    `{'a': {'b': 2, 'b_path': 'config2.yaml'}` If you replace '@merge_after'
    by '@merge_before', it will be `{'a': {'b': 1, 'b_path': 'config2.yaml'}`
    instead.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = -20.0

    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,
    ) -> Dict[str, Any]:
        """Pre-merge processing."""
        items = list(flat_dict.items())
        for flat_key, val in items:
            if flat_key.endswith('@merge_after'):
                # NOTE: we allow new keys here fore more flexibility but we note that
                # the merge following this pre-merge will avoid the creation of
                # new keys if needed.
                if not isinstance(val, str) or not val.endswith('.yaml'):
                    raise ValueError(
                        "Key with '@merge_after' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the dict
                flat_dict[flat_key[:-len('@merge_after')]] = val
                del flat_dict[flat_key]
                # Merge + process the dicts
                flat_dict = merge_flat_paths_processing(
                    flat_dict,
                    val,
                    processing_list=processing_list,
                    allow_new_keys=True,
                    preprocess_first=False,  # Already processed
                )
            if flat_key.endswith('@merge_before'):
                # NOTE: we allow new keys here fore more flexibility but we note that
                # the merge following this pre-merge will avoid the creation of
                # new keys if needed.
                if not isinstance(val, str) or not val.endswith('.yaml'):
                    raise ValueError(
                        "Key with '@merge_before' tag must be associated "
                        "to a string corresponding to a *yaml* file."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag in the key
                flat_dict[flat_key[:-len('@merge_before')]] = val
                del flat_dict[flat_key]
                # Merge + process the dicts
                flat_dict = merge_flat_paths_processing(
                    val,
                    flat_dict,
                    processing_list=processing_list,
                    allow_new_keys=True,
                    preprocess_second=False,  # Already processed
                )
        return flat_dict


class ProcessCopy(Processing):
    """Copy a value with '@copy' tag.

    Tag your key with '@copy' and with value the name of the flat key to copy.
    This processing is a pre-merge processing only and occurs after most of
    the other pre-merge processing.
    Pre-merge order: 20.0

    Examples
    --------
    .. code-block:: yaml

        # config.yaml
        a:
          b: 1
          c@copy: a.b

    Before merging, the config is interpreted as the dict
    `{'a': {'b': 1, 'c': 1}}`.
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = 20.0

    # pylint: disable=unused-argument
    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-merge processing."""
        items = list(flat_dict.items())
        for flat_key, val in items:
            if flat_key.endswith('@copy'):
                if not isinstance(val, str):
                    raise ValueError(
                        "Key with '@copy' tag must be associated "
                        "to a string corresponding to a flat key."
                        f"The problem occurs at key: {flat_key}"
                    )
                # Remove the tag and update the dict
                flat_dict[flat_key[:-len('@copy')]] = flat_dict[val]
                del flat_dict[flat_key]
        return flat_dict


class ProcessProtect(Processing):
    """Make a value protected with '@protect' tag.

    The protected value cannot be changed on merge as long as this processing
    is used. Otherwise, it raises an error. This processing store protected
    keys and values and remove the '@protect' tag. It always occurs before anything
    else on pre-merge processing and after everything else on post-merge.
    Pre-merge order: -40.0
    Post-merge order: 40.0
    """

    def __init__(self) -> None:
        super().__init__()
        self.premerge_order = -40.0
        self.postmerge_order = 40.0
        self.protected_dict: Dict[str, Any] = {}

    # pylint: disable=unused-argument
    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-merge processing."""
        items = list(flat_dict.items())
        for flat_key, val in items:
            if flat_key.endswith('@protect'):
                new_key = flat_key[:-len('@protect')]
                if new_key in flat_dict and self.protected_dict[new_key] != val:
                    raise ValueError(
                        f"Attempt to change a protected key (tagged by '@protect'): "
                        f"{new_key}. Protected value: {self.protected_dict[new_key]}, "
                        f"new value: {val}."
                    )
                # Remove the tag in the key
                flat_dict[new_key] = val
                del flat_dict[flat_key]
                # Store the protected key and value
                self.protected_dict[flat_key[:-len('@protect')]] = val
        return flat_dict

    # pylint: disable=unused-argument
    def postmerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Post-merge processing."""
        for key, val in self.protected_dict.items():
            if key not in flat_dict:
                raise ValueError(
                    f"Protected key (tagged by '@protect') {key} has been removed "
                    f"after merge. Protected value: {val}, new value: nothing."
                )

            if flat_dict[key] != val:
                raise ValueError(
                    f"Protected key (tagged by '@protect') {key} has been changed "
                    f"after merge. Protected value: {val}, new value: {flat_dict[key]}."
                )
        return flat_dict



"""Routines to manipulate dictionaries with processing."""
from typing import Any, Dict, List, Union

from cliconfig.dict_routines import _flat_before_merge, load_dict, merge_flat, save_dict


class Processing():
    """Processing base class."""

    def __init__(self) -> None:
        self.premerge_order = 0.0
        self.postmerge_order = 0.0
        self.presave_order = 0.0
        self.postload_order = 0.0

    # pylint: disable=unused-argument
    def premerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-merge processing.

        Function that can be applied to the flat dict to modify it
        before merging . It takes a flat dict and returns a flat dict.
        """
        return flat_dict

    # pylint: disable=unused-argument
    def postmerge(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Post-merge processing.

        Function that can be applied to the flat dict to modify it
        after merging . It takes a flat dict and returns a flat dict.
        """
        return flat_dict

    # pylint: disable=unused-argument
    def presave(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Pre-save processing.

        Function applied to the flat dict to modify it before
        saving it. It takes a flat dict and returns a flat dict.
        """
        return flat_dict

    # pylint: disable=unused-argument
    def postload(
        self,
        flat_dict: Dict[str, Any],
        processing_list: list,  # noqa
    ) -> Dict[str, Any]:
        """Post-load processing.

        Function applied to the flat dict to modify it after
        loading it. It takes a flat dict and returns a flat dict.
        """
        return flat_dict


def merge_flat_processing(
    dict1: Dict[str, Any],
    dict2: Dict[str, Any],
    processing_list: List[Processing],
    *,
    allow_new_keys: bool = True,
    preprocess_first: bool = True,
    preprocess_second: bool = True,
    postprocess: bool = True,
) -> Dict[str, Any]:
    """Flatten, merge dict2 into dict1 and apply pre and post processing.

    Similar to :func:`cliconfig.dict_routines.merge_flat` but with processing
    applied before and/or after the merge. Work even if the dicts have
    a mix of nested and flat dictionaries.

    Parameters
    ----------
    dict1 : Dict[str, Any]
        The first dict. It can be nested, flat or a mix of both.
    dict2 : Dict[str, Any]
        The second dict to merge into dict1.
    processing_list: List[Processing]
        The list of processing to apply during the merge. Only premerge and
        postmerge methods are applied. The order of the processing is given
        by the premerge_order and postmerge_order attributes of the processing.
    allow_new_keys : bool, optional
        If True, new keys (that are not in dict1) are allowed in dict2.
        By default True.
    preprocess_first : bool, optional
        If True, apply pre*merge processing to dict1. By default True.
    preprocess_second : bool, optional
        If True, apply pre*merge processing to dict2. By default True.
    postprocess : bool, optional
        If True, apply post*merge processing to the merged dict. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and dict2 has new keys that are not in dict1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.
        See last example. You may consider calling :func:`clean_pre_flat` on the input
        dicts in that case.

    Returns
    -------
    flat_dict : Dict[str, Any]
        The flat dict (all keys are at the root and separated by dots).
    """
    # Get the pre-merge and post-merge order
    pre_order_list = sorted(processing_list, key=lambda x: x.premerge_order)
    post_order_list = sorted(processing_list, key=lambda x: x.postmerge_order)
    # Flatten the dictionaries
    flat_dict1, flat_dict2 = _flat_before_merge(dict1, dict2)
    # Apply the premerge processing
    for processing in pre_order_list:
        if preprocess_first:
            flat_dict1 = processing.premerge(flat_dict1, processing_list)
        if preprocess_second:
            flat_dict2 = processing.premerge(flat_dict2, processing_list)
    # Merge the dictionaries
    flat_dict = merge_flat(flat_dict1, flat_dict2, allow_new_keys=allow_new_keys)
    # Apply the postmerge processing
    for processing in post_order_list:
        if postprocess:
            flat_dict = processing.postmerge(flat_dict, processing_list)
    return flat_dict


def merge_flat_paths_processing(
    dict_or_path1: Union[str, Dict[str, Any]],
    dict_or_path2: Union[str, Dict[str, Any]],
    processing_list: List[Processing],
    *,
    allow_new_keys: bool = True,
    preprocess_first: bool = True,
    preprocess_second: bool = True,
    postprocess: bool = True,
) -> Dict[str, Any]:
    """Flatten, merge and apply processing to two dictionaries or their yaml paths.

    Similar to :func:`cliconfig.dict_routines.merge_flat_paths` but with processing
    applied before and/or after the merge. It is also similar to
    :func:`merge_flat_processing` but allows to pass dictionaries or yaml paths.
    Work even if the dicts have a mix of nested and flat dictionaries.

    Parameters
    ----------
    dict_or_path1 : Union[str, Dict[str, Any]]
        The first dict or its path.
    dict_or_path2 : Union[str, Dict[str, Any]]
        The second dict or its path, to merge into first dict.
    processing_list: List[Processing]
        The list of processing to apply during the merge. Only premerge and
        postmerge methods are applied. The order of the processing is given
        by the premerge_order and postmerge_order attributes of the processing.
    allow_new_keys : bool, optional
        If True, new keys (that are not in dict1) are allowed in dict2.
        By default True.
    preprocess_first : bool, optional
        If True, apply pre-merge processing to dict1. By default True.
    preprocess_second : bool, optional
        If True, apply pre-merge processing to dict2. By default True.
    postprocess : bool, optional
        If True, apply post-merge processing to the merged dict. By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and dict2 has new keys that are not in dict1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.
        See last example. You may consider calling :func:`clean_pre_flat` on the input
        dicts in that case.

    Returns
    -------
    flat_dict : Dict[str, Any]
        The flat dict (all keys are at the root and separated by dots).
    """
    dicts = []
    for dict_or_path in [dict_or_path1, dict_or_path2]:
        if isinstance(dict_or_path, str):
            _dict = load_dict(dict_or_path)
        else:
            _dict = dict_or_path
        dicts.append(_dict)
    dict1, dict2 = dicts[0], dicts[1]
    flat_dict = merge_flat_processing(
        dict1,
        dict2,
        processing_list,
        allow_new_keys=allow_new_keys,
        preprocess_first=preprocess_first,
        preprocess_second=preprocess_second,
        postprocess=postprocess,
    )
    return flat_dict


def save_processing(
    in_dict: Dict[str, Any],
    path: str,
    processing_list: List[Processing]
) -> None:
    """Save a dict and apply pre-save processing before saving.

    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to save.
    path : str
        The path to the yaml file to save the dict.
    processing_list: List[Processing]
        The list of processing to apply before saving. Only presave
        method is applied. The order of the processing is given
        by the presave_order attribute of the processing.
    """
    # Get the pre-save order
    order_list = sorted(processing_list, key=lambda x: x.presave_order)
    # Apply the pre-save processing
    for processing in order_list:
        in_dict = processing.presave(in_dict, processing_list)
    # Save the dict
    save_dict(in_dict, path)


def load_processing(path: str, processing_list: List[Processing]) -> Dict[str, Any]:
    """Load a dict from yaml file and apply post-load processing.

    Parameters
    ----------
    path : str
        The path to the file to load the dict.
    processing_list: List[Processing]
        The list of processing to apply after loading. Only postload
        method is applied. The order of the processing is given
        by the postload_order attribute of the processing.

    Returns
    -------
    out_dict: Dict[str, Any]
        The loaded dict.
    """
    # Get the post-load order
    order_list = sorted(processing_list, key=lambda x: x.postload_order)
    # Load the dict
    out_dict = load_dict(path)
    # Apply the post-load processing
    for processing in order_list:
        out_dict = processing.postload(out_dict, processing_list)
    return out_dict



"""Functions to manipulate config as dict with yaml files and CLI."""
import sys
from typing import Any, Dict, List, Optional

from cliconfig.cli_parser import parse_cli
from cliconfig.dict_routines import unflatten
from cliconfig.process_routines import (
    Processing,
    load_processing,
    merge_flat_paths_processing,
    merge_flat_processing,
)


def make_config(
    *default_config_paths: str,
    processing_list: Optional[List[Processing]] = None,
) -> Dict[str, Any]:
    r"""Make a config from default configs and CLI arguments and apply processing.

    It uses the CLI Config routines to parse the CLI arguments and merge
    them in a simple way. Apply the processing functions on each merge.

    Parameters
    ----------
    default_config_paths : Tuple[str]
        Paths to default configs. They are merged in order and new keys
        are allowed.
    processing_list: Optional[List[Processing]], optional
        The list of processing to apply during each merge. Only premerge and
        postmerge methods are applied. The order of the processing is given
        by the premerge_order and postmerge_order attributes of the processing.
        If None, no processing is applied. By default None.

    Raises
    ------
    ValueError
        If allow_new_keys is False and CLI has new keys that are not
        in default configs.

    Returns
    -------
    config : Dict[str, Any]
        The merged config.

    Examples
    --------
    ::

        # main.py
        config = make_config('data.yaml', 'model.yaml', 'train.yaml')

    .. code-block:: text

        python main.py -- config bestmodel.yaml,mydata.yaml \
            --architecture.layers.hidden_dim 64

    """
    config: Dict[str, Any] = {}
    config_paths, config_cli_params = parse_cli(sys.argv)
    print(
        f"[CONFIG] Merge {len(default_config_paths)} default configs, "
        f"{len(config_paths)} additional configs and "
        f"{len(config_cli_params)} CLI parameter(s)."
    )

    if processing_list is None:
        processing_list = []

    for default_config_path in default_config_paths:
        # Allow new keys for default configs
        config = merge_flat_paths_processing(
            config,
            default_config_path,
            processing_list,
            allow_new_keys=True,
            preprocess_first=False,  # Already processed
        )

    for config_path in config_paths:
        # Disallow new keys for additional configs
        config = merge_flat_paths_processing(
            config,
            config_path,
            processing_list,
            allow_new_keys=False,
            preprocess_first=False,
        )
    # Disallow new keys for CLI parameters
    config = merge_flat_processing(
        config,
        config_cli_params,
        processing_list,
        allow_new_keys=False,
        preprocess_first=False)
    # Unflatten the config
    config = unflatten(config)
    return config


def load_config(
    path: str,
    default_config_paths: Optional[List[str]] = None,
    processing_list: Optional[List[Processing]] = None,
) -> Dict[str, Any]:
    """Load config from a file and merge into optional default configs.

    Parameters
    ----------
    path : str
        The path to the file to load the configuration.
    default_config_paths : Optional[List[str]], optional
        Paths to default configs. They are merged in order, new keys are allowed.
        Then, the loaded config is merged into the result. None for no default configs.
        By default None.
    processing_list: Optional[List[Processing]]
        The list of processing to apply after loading. Only postload
        method is applied. The order of the processing is given
        by the postload_order attribute of the processing.
        If None, no processing is applied. By default None.

    Returns
    -------
    config: Dict[str, Any]
        The loaded config.
    """
    if processing_list is None:
        processing_list = []
    config: Dict[str, Any] = {}
    if default_config_paths:
        for config_path in default_config_paths:
            config = merge_flat_paths_processing(
                config,
                config_path,
                processing_list=processing_list,
                allow_new_keys=True,
                preprocess_first=False,  # Already processed
            )
    # Disallow new keys for loaded config
    loaded_config = load_processing(path, processing_list)
    config = merge_flat_processing(
        config,
        loaded_config,
        processing_list=processing_list,
        allow_new_keys=False,
        preprocess_first=False,
    )
    # Unflatten the config
    config = unflatten(config)
    return config



"""Tests for dict routines with preprocessing."""
import shutil

import pytest_check as check
import yaml

from cliconfig.process_routines import (
    Processing,
    load_processing,
    merge_flat_paths_processing,
    merge_flat_processing,
    save_processing,
)
from tests.conftest import ProcessAdd1, ProcessKeep


def test_processing() -> None:
    """Test Processing."""
    in_dict = {"a.b": 1, "b": 2, "c.d.e": 3, "c.d.f": [2, 3]}
    base_process = Processing()
    check.equal(
        (
            base_process.premerge(in_dict, []),
            base_process.postmerge(in_dict, []),
            base_process.presave(in_dict, []),
            base_process.postload(in_dict, []),
        ),
        (in_dict, in_dict, in_dict, in_dict),
    )


def test_merge_flat_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test merge_flat_processing."""
    dict1 = {"a": {"b": 1, "c": 2, "d@keep": 3}}
    dict2 = {"a": {"b": 2, "c@add+1": 3, "d": 4}}
    flat_dict = merge_flat_processing(
        dict1,
        dict2,
        processing_list=[process_add1, process_keep],
        allow_new_keys=False,
    )
    check.equal(
        flat_dict,
        {"a.b": 2, "a.c": 4, "a.d": 3},
    )
    check.equal(process_keep.keep_vals, {})


def test_merge_flat_paths_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test merge_flat_paths_processing."""
    dict1 = {"param1@add+1": 0, "param2.param3@keep": 1}
    dict2 = {"param2.param3": 3}
    expected_dict = {"param1": 1, "param2.param3": 1}
    check.equal(
        merge_flat_paths_processing(
            "tests/configs/configtag1.yaml",
            "tests/configs/configtag2.yaml",
            [process_add1, process_keep],
            allow_new_keys=False,
        ),
        expected_dict,
    )
    check.equal(
        merge_flat_paths_processing(
            dict1,
            dict2,
            [process_add1, process_keep],
            allow_new_keys=False,
        ),
        expected_dict,
    )
    check.equal(process_keep.keep_vals, {})


def test_save_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test save_processing."""
    in_dict = {"param1@add+1": 0, "param2.param3@add+1": 1}
    save_processing(in_dict, "tests/tmp/config.yaml", [process_add1, process_keep])
    with open("tests/tmp/config.yaml", "r", encoding="utf-8") as yaml_file:
        loaded_dict = yaml.safe_load(yaml_file)
    check.equal(loaded_dict, {"param1": 1, "param2.param3": 2})
    check.equal(process_keep.keep_vals, {})
    shutil.rmtree("tests/tmp")


def test_load_processing(
    process_add1: ProcessAdd1,
    process_keep: ProcessKeep,
) -> None:
    """Test load_processing."""
    process_keep.keep_vals = {"param2.param3": 0}
    config = load_processing(
        "tests/configs/configtag2.yaml",
        [process_add1, process_keep],
    )
    check.equal(config, {"param2.param3": 0})
